// ***********************************************************************
// VISUAL MARRIAGE Applet
//
// Visualizes the algorithm for finding a man-optimal stable marriage.
// The simulated algorithm is the one of Gale and Shapley.
//
// by Rolf Bruderer, Oct 2003
// ***********************************************************************

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.BorderFactory; 
import javax.swing.border.*;
import javax.swing.Box.Filler;
import java.util.*;


class MainWindow
	// Main-Window for Visual Marriage Applet.
	// Contains the whole GUI.
	// Initialize and manage the most important objects.

extends 
	JFrame

implements 
	ActionListener
	
{	
	private JPanel main_panel;
		// all GUI components will be in this main_panel
		// such that resizing the window does not affect the layout.

	private JPanel control_panel;
		// All control buttons will be in this panel.

	private World world;
		// Visualized world with men and women.
	
	private JPanel world_panel;
		// panel containing the world.
	
	
	private JSplitPane split_pane;
		// split pane for resizing the height of tables.

	private Thread world_engine;  
		// Thread for always keeping world running.
	
	
	private PersonTableModel men, women;
		// women and men.

	
	private PersonTable men_table, women_table;
		// tables visualizing women and men.
		
	private JButton start_button, stop_button, step_button, reset_button;
		// buttons for controling the animation.
		
	private JButton history_button, options_button;
		// buttons to open the options and history windows.
	
	private ButtonGroup preferences_option_group;
		// A button group to specify how to visualize preferences lists.
	
	
	private JRadioButton symbol_option, name_option;
		// radio buttons to choose how to visualize preference lists.
	
	private OptionsWindow options_window;
		// the frame window to control all options.

	private HistoryWindow history_window;
		// the frame window to log the history.
		
	public MainWindow()
	// create and initialize frame.
	{

		// create GUI.
		buildGUI (getContentPane());
		
		// create options_window.
		options_window = new OptionsWindow(this);
		options_window.setVisible(false);
		
		// create history_window.
		history_window = new HistoryWindow();
		history_window.setVisible(false);
		Person.history_window = history_window;
		
		// create and add ComponentResizeListener to get informed when window resizes.
		addComponentListener(new ComponentResizeListener(this));
		
		// cretae and add WindowCloseListener, 
		// to close other related windows when this main window closes.
		addWindowListener(new WindowCloseListener());
				
		// create empty vectors men_names and women_names.
		men_names = new Vector();
		women_names = new Vector();		
		
		// create persons.
		createPersons(9);
		
		// create and start `world_engine' which lets the world run.
		world_engine = new Thread (world);
		world_engine.setPriority(Thread.MIN_PRIORITY);
		world_engine.start ();	
				
  	}

	private JPanel getASeparator() 
	// create and return an empty JPanel to separate buttons in control_panel.
	{
		JPanel separator = new JPanel();
		separator.setBackground(Color.black);		
		separator.setPreferredSize(new Dimension(75, 10));		
		return separator;
	}
	
	private JButton getAButton(String text) 
	// create and return an empty JPanel to separate buttons in control_panel.
	{	
		Dimension button_size = new Dimension(75, 25);
		Insets button_margin = new Insets(0,0,0,0);	
		Color button_color = new Color(64, 64, 192);	
		JButton button = new JButton (text);
		button.setBackground(button_color);
		button.setForeground(Color.white);
		button.setMargin(button_margin);
		button.setPreferredSize(button_size);
		button.setActionCommand(text);
		button.addActionListener(this);
		return button;
	}

    private void buildGUI(Container container) 
	// Build main parts of the GUI.
	{	
		// set options.
		setBackground (Color.black);
		container.setBackground (Color.black); 
		container.setLayout (new FlowLayout (FlowLayout.CENTER));
		setTitle (" V I S U A L   M A R R I A G E");
		setVisible (true);
		
		// create main_panel.
		main_panel = new JPanel();
		main_panel.setBackground(Color.black); 
		main_panel.setLayout (new FlowLayout (FlowLayout.CENTER));		

		// create nice border for panels.
		Border panel_border = BorderFactory.createEtchedBorder(EtchedBorder.RAISED, Color.gray, Color.darkGray);

		// create control_panel.
		control_panel = new JPanel();
		control_panel.setBackground(Color.black);
		control_panel.setLayout(new FlowLayout (FlowLayout.CENTER));
		control_panel.setBorder(panel_border);		

		// create world.
		world = new World ();
				
		// create world_panel.
		world_panel = new JPanel();
		world_panel.setBackground(Color.black);
		world_panel.setLayout(new FlowLayout (FlowLayout.CENTER));
		world_panel.setBorder(panel_border);						
		
		// create buttons.
		start_button = getAButton("start");
		stop_button = getAButton("stop");
		step_button = getAButton("step");
		reset_button = getAButton("reset");
		history_button = getAButton("History");
		options_button = getAButton("Options");
		
		/*
		Font button_font = history_button.getFont();
		button_font = button_font.deriveFont(button_font.size()-4);
		history_button.setFont(button_font);
		*/		
		
		// create person tables and their views.
		men = new PersonTableModel ();
		men_table = new PersonTable (men, this);
		women = new PersonTableModel ();
		women_table = new PersonTable (women, this);
				
		// connect women and men tables.
		women.addTableModelListener(men_table.getTable ());
		men.addTableModelListener(women_table.getTable ());
		
		// create split_pane with women and men tables.
		split_pane = new JSplitPane(JSplitPane.VERTICAL_SPLIT, men_table, women_table);
		split_pane.setBackground (Color.black);
		split_pane.setResizeWeight (0.5);
		
		// lay out everything & set size.
		world_panel.add(world);
		main_panel.add(world_panel);
		control_panel.add(start_button);		
		control_panel.add(step_button);
		control_panel.add(stop_button);
		control_panel.add(getASeparator());
		control_panel.add(getASeparator());
		control_panel.add(reset_button);
		control_panel.add(getASeparator());
		control_panel.add(getASeparator());
		control_panel.add(getASeparator());
		control_panel.add(getASeparator());
		control_panel.add(getASeparator());
		control_panel.add(history_button);
		control_panel.add(options_button);
		
		main_panel.add(control_panel);
		main_panel.add (split_pane);		
		container.add (main_panel);
		int worldPanelWidth = 580 - 98;
		int worldPanelHeight  = 300;
		world_panel.setPreferredSize (new Dimension (worldPanelWidth, worldPanelHeight));
		world.setSize (new Dimension (worldPanelWidth - 10, worldPanelHeight - 10));
		split_pane.setPreferredSize (new Dimension (580, 800 - worldPanelHeight - 80)); 		
		main_panel.setPreferredSize (new Dimension (600, 800));		
		control_panel.setPreferredSize (new Dimension (90, worldPanelHeight));
		setSize(new Dimension(600,800));
		
  	}
	
	private void distributeWomen() 
	// will distribute women at regular intervals in world space.
	{


		// are there any women to distribute?
		int nofWomen = women.getRowCount();
		if (nofWomen == 0) {
			// nothing to do, just return.
			return;
		}

		// find optimal number of rows/columns to distribute womans regularly.
		int w = world.getWidth ();
		int h = world.getHeight () - 20;
		double opt_bal = 0; 
		int opt_rows = 0;
		int cell_width = 1;
		int cell_height = 1;		
		double bal = 0.0;
		int rows = 0;		
		while (bal < 1.0) {
		
			// take last calculated as new optimum.
			opt_bal = bal;
			opt_rows = rows;
			
			// take one row more.
			rows++;
			
			// calculate number of columns and rest.
			int columns = nofWomen / rows;
						
			// calculate side balance of one cell.
			if (columns > 0) {
				cell_width = w / columns;
				cell_height = h / rows;
				bal = (double)cell_width / (double) cell_height;
			}
			else {
				bal = 2.0;
			}
			
		}
		
		// which is better? rows or opt_rows?
		bal = (double)cell_height / (double) cell_width;
		if (bal > opt_bal) {
			opt_rows = rows;
			opt_bal = bal;
		}
		
		// now distribute all women regularly to opt_rows rows.
		double womenPerRow = ((double)nofWomen) / (double)opt_rows;
		cell_height = h / opt_rows;
		int y = cell_height / 2;
		double progress = 0.0;
		int distributedWomen = 0;
		Enumeration enum = women.elements();
		while (enum.hasMoreElements()) {
		
			// increase progress.
			progress = progress + womenPerRow;					

			// calculate properties of this row.
			int womenInRow = (int)progress - distributedWomen;			
			if (progress + 0.5 > (double)nofWomen) {
				// ensure to stop.
				womenInRow = nofWomen - distributedWomen;
				progress = (double)nofWomen;
			}
			cell_width = w / womenInRow;
			int x = cell_width / 2;
			
			while (distributedWomen < (int)progress) {
			
				Person p = (Person)enum.nextElement ();
				p.setPosition (x, y);
				x = x + cell_width;
				distributedWomen++;
								
			}
			
			y = y + cell_height;
		}
		
	}
		
	public void paint(Graphics g) {
	
		Graphics2D g2d = (Graphics2D)g;		
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		super.paint(g2d);		
		
	}	
	
	public boolean is_reseted;
	// Is the window in initialized (not running) state?
	// (important for dragging preferences, not allowed when not reseted)

	public void reset() 
	// stop world and reset all man to restart searching for a woman.
	{
	
		world.stop();
		
		history_window.clear();
		
		// set all men to their start position and reset them.
		Enumeration enum = men.elements();
		int i = 0;
		while (enum.hasMoreElements()) {	
			Man man = (Man)enum.nextElement();
			man.setPosition(-25 * i, world.getHeight() / 2);
			man.reset();			
			i++;
		}

		// reset all women.
		enum = women.elements();
		while (enum.hasMoreElements()) {	
			Woman woman = (Woman)enum.nextElement();
			woman.reset();
		}

		// update views.
		men.fireTableDataChanged();
		women.fireTableDataChanged();		
		world.repaint();
	}	
	
	public void bringWindowsToFront() 
	// shows the main window 
	// and brings all other showing windows of the application to front.
	{
		setLocation(getX(), getY());
		show();
		setState(NORMAL);
		if (options_window.isShowing()) {
			options_window.show();
			options_window.setState(options_window.NORMAL);
			options_window.validate();
			options_window.repaint();
		}
		if (history_window.isShowing()) {
			history_window.show();
			history_window.setState(history_window.NORMAL);		
			history_window.validate();
			history_window.repaint();
		}
		
	}

	private void showOptionsWindow() 
	// show the window to control some options.
	{
		int y = (int)options_button.getLocationOnScreen().getY();
		Point point = new Point(getX() + getWidth() - options_window.getWidth(), y + 5);
		options_window.setLocation(point);
		options_window.show();
		options_window.setState(options_window.NORMAL);
		options_window.validate();
		options_window.repaint();
	}
	
	private void showHistoryWindow() 
	// show the window to view the history of marriages and divorces.
	{
		int y = (int)history_button.getLocationOnScreen().getY();
		Point point = new Point(getX() + getWidth() - history_window.getWidth(), y + 5);
		history_window.setLocation(point);
		history_window.show();
		history_window.setState(history_window.NORMAL);		
		history_window.validate();
		history_window.repaint();
	}
	
	public void showNamesOnlyInTables() {
		men_table.showNamesOnly();
		women_table.showNamesOnly();	
	}
	
	public void showSymbolsOnlyInTables() {
		men_table.showSymbolsOnly();
		women_table.showSymbolsOnly();	
	}
	
	public void showSymbolsAndNamesInTables() {
		men_table.showSymbolsAndNames();
		women_table.showSymbolsAndNames();	
	}
	
	public void setNamesVisibleForMarriedPersons(boolean flag) {
		Person.setNamesVisibleForMarriedPersons(flag);
		world.repaint();
	}
	
	public void setNamesVisibleForActingPersons(boolean flag) {
		Person.setNamesVisibleForActingPersons(flag);
		world.repaint();		
	}

	private Vector men_names, women_names;
		// names given so far to mens.	

	private void backupPersonNames() 
	// backup all given names to 'men_names' and 'women_names' for later reuse.
	{
		int index = 0;
		Enumeration e = men.elements();		
		while (e.hasMoreElements()) {
			Person p = (Person)e.nextElement();
			if (index < men_names.size()) {
				men_names.setElementAt(p.getName(), index);
			}
			else {
				men_names.add(index, p.getName());
			}			
			index++;					
		}
		index = 0;
		e = women.elements();		
		while (e.hasMoreElements()) {
			Person p = (Person)e.nextElement();
			if (index < women_names.size()) {
				women_names.setElementAt(p.getName(), index);
			}
			else {
				women_names.add(index, p.getName());
			}			
			index++;					
		}

	}

	private String getWomanName(int index) 
	// get name for creation of a woman at index 'index'.
	// either generate a new default name
	// or reuse former backuped name.
	{	
		if (index < women_names.size()) {
			return (String)women_names.get(index);
		}
		else {
			return new String("Woman " + String.valueOf(index + 1));
		}
	}

	private String getManName(int index) 
	// get name for creation of a man at index 'index'.
	// either generate a new default name 
	// or reuse former backuped name.	
	{	
		if (index < men_names.size()) {
			return (String)men_names.get(index);
		}
		else {
			return new String("Man " + String.valueOf(index + 1));
		}
	}

	public void createPersons(int count) 
	// clear existing persons and reinitialize everything with the given number of persons.	
	{
		
		// before clearing old persons, backup all given names, to reuse them later.
		backupPersonNames();
		
		// first stop the world and empty the world and the tables.
		world.stop();
		world.removeAllPersons();
		men.removeAllElements();
		women.removeAllElements();		
		history_window.clear();		

		// choose size of persons corresponding to count.
		double size = world.getWidth ()/(5*Math.sqrt((float)count));
		if (size > 15.0) {   // Maximum size
			size = 15.0;
		}
		if (size < 3.0) {   // Minimum size
			size = 3.0;				
		}
		Person.setSize(size);
				
		// Create color factory to produce 
		// diffrent colors for each person.
		ColorFactory colors = new ColorFactory(count);

		// create women.
		for (int i = 0; i < count; i++) {
			Woman woman = new Woman ();			
			woman.setColor(colors.getNextColor());
			woman.setName(getWomanName(i));
			woman.setTableRow(women, i);
			world.add(woman);
			women.add(woman);			
		}

		// create men.
		for (int i = 0; i < count; i++) {
			Man man = new Man (-25 * i, world.getHeight() / 2);
			man.setColor(colors.getNextColor());
			man.setName (getManName(i));
			man.setTableRow(men, i);
			world.add (man);
			men.add (man);
		}
		
		// create preferences.
		for (int i = 1; i <= count; i++) {
			men.extend_preferences (women.elementAt (i-1));			
			women.extend_preferences (men.elementAt (i-1));
		}		
		
		// randomize preferences.
		for (int i = 0; i < men.getRowCount(); i++) {
			men.elementAt(i).randomizePreferences ();			
			women.elementAt(i).randomizePreferences ();
		}

		// distribute women regularly.
		distributeWomen();
		
		// update everything.
		validateTree();
		men_table.adjustPreferenceColumnWidth ();
		women_table.adjustPreferenceColumnWidth ();
		world.repaint();
		repaint();		
		
	}
	
	public void actionPerformed(ActionEvent e) {
	
	    if (e.getActionCommand().equals("start")) {
			is_reseted = false;
	    	world.start();	  
	    }
	    if (e.getActionCommand().equals("stop")) {
	    	world.stop();	   
	    }
	    if (e.getActionCommand().equals("step")) {
			is_reseted = false; 
	    	world.step();	    
	    }
	    if (e.getActionCommand().equals("reset")) {
	    	reset();    
			is_reseted = true; 
	    }

		if (e.getActionCommand().equals("Options")) {
	    	showOptionsWindow();
	    }
		
		if (e.getActionCommand().equals("History")) {
	    	showHistoryWindow();
	    }
   	}
	
	class 
		ComponentResizeListener	
	extends 
		ComponentAdapter	
	{
	
		private MainWindow this_win;
	
		public ComponentResizeListener(MainWindow window) 
		// initialize the resize listener, with the window it is listening to.
		{
			this_win = window;
		}
	
		public void componentResized(ComponentEvent e) 
		// Resize the 'split_pane' together with this window.
		{
				
			Dimension size = this_win.getSize();
			
			int width = (int) this_win.getSize().getWidth();
			int height = (int) this_win.getSize().getHeight(); 
			
					
			if (width < 600) {
				size.setSize(600, height);
			}						
			main_panel.setPreferredSize (size);
			main_panel.setSize (size);		
			this_win.setSize (size);
			
			split_pane.setPreferredSize (new Dimension (width - 20, height - world_panel.getHeight() - 80));
			split_pane.setSize (new Dimension (width - 20, height - world_panel.getHeight() - 80));	
			
			this_win.validateTree();			

			// adjust preference column width in tables.
			men_table.adjustPreferenceColumnWidth ();
			women_table.adjustPreferenceColumnWidth ();
			
		}		
	
	}		
	
	class 
		WindowCloseListener	
	extends 
		WindowAdapter	
	{
	
		public void windowClosing(WindowEvent e) 
		// Close 'history_window' and 'options_window' when the MainWindow is closed.
		{
			history_window.dispose();
			options_window.dispose();
			world.destroy();
		}
		
	}
	
	
}
