// ***********************************************************************
// VISUAL MARRIAGE Applet
//
// Visualizes the algorithm for finding a man-optimal stable marriage.
// The simulated algorithm is the one of Gale and Shapley.
//
// by Rolf Bruderer, Oct 2003
// ***********************************************************************

import java.awt.*;
import javax.swing.*;
import javax.swing.table.*;
import java.util.Random;


public abstract class Person
	// Persons that can be visualized in a World and in a PersonTable.
	// (Man and Woman will extend this class)
	
implements Icon
	
{ 

	static boolean pendingAction;
	// is there at least one person in acting state?
	// if this is true, only persons that are part of this action (not in normal moving state)
	// are allowed to move.	
	
	public static HistoryWindow history_window;
	// the history window, where all person actions should be logged in.

	double x, y;
	// the person's position in the x/y-plane of the world.
	
	Color color;
	// the person's color.

	String name;
	// name of the person (or just an id).
	
	boolean married;
	// is this person married?
		
	Person partner;
	// the person with which this person is married.	

	PreferenceList preferences;
	// preferences of the person to possible partners.
	
	JComponent symbol_component;
	// component that can be used to paint the person.
	
	JComponent name_component;
	// component that can be used to paint the name of the person.
	
	JComponent symbol_with_name_component;
	// component that can be used to paint the symbol of the person together with the name.
	
	PersonTableModel ownerTable;
	// the table where this person appears in (as a row).
	
	int ownerTableRow;
	// the row in 'ownerTable' that contains this person.
	
	static Random rnd;
	// randomizer for randomizing preferences.	
		
	static double size = 15.0;
	// width and length of the image visualizing a person.
	
	static double speed = 1.0;
	// speed of a person. (default value is 1.0).	
	
	static double slow_speed = 0.25;
	// slow motion speed of a person. (default value is 4 times slower than normal speed).
	
	public Person () 
	// initialize person.
	// choose color with color factory.
	{

		// initialize random generator, if neccessary.
		if (rnd == null) {
			rnd = new Random();
		}

		// initialize important instance fields.
		x = 0;
		y = 0;
		married = false;
		partner = null;	
		color = new Color(0, 0, 0);
		preferences = new PreferenceList ();
		buildPaintComponents ();				
	}
	
	public void reset()	
	// reset a persons state.
	{
		married = false;	
		partner = null;	
		preferences.clearHighlightColors();
		preferences.clearBackgroundColors();
	}	

	
	
	public void buildPaintComponents () 
	// create components to paint the name and the symbol of this person.
	{
		// create name_component.
		JLabel name_label = new JLabel(name);
		name_label.setVerticalAlignment (JLabel.CENTER);
		name_label.setHorizontalAlignment (JLabel.CENTER);
		name_label.setFont (name_label.getFont().deriveFont(Font.PLAIN));
		name_label.setBackground(Color.black);
		name_label.setForeground(Color.white);
		name_label.setOpaque(true);
		name_label.setLocation(new Point(0,2));
		name_component = name_label;

		// create symbol_component.
		JLabel symbol_label = new JLabel ((Icon) this);		
		symbol_label.setHorizontalAlignment (JLabel.CENTER);
		symbol_label.setVerticalAlignment (JLabel.CENTER);
		symbol_label.setBackground(Color.black);
		symbol_label.setOpaque(true);
		symbol_component = symbol_label;
		
		// create symbol_with_name_component.
		JLabel symbol_with_name_label = new JLabel(name, (Icon) this, JLabel.CENTER);
		symbol_with_name_label.setVerticalAlignment (JLabel.CENTER);
		symbol_with_name_label.setFont (name_label.getFont().deriveFont(Font.PLAIN));
		symbol_with_name_label.setBackground(Color.black);
		symbol_with_name_label.setForeground(Color.white);
		symbol_with_name_label.setOpaque(true);
		symbol_with_name_component = symbol_with_name_label;	
		
	}
	
		
	static void setSize (double s)
	// set width and length of the image visualizing a person, for all persons.
	{
		size = s;
	}
	
	public void setPosition (int xp, int yp) 
	// set position of this person to `xp' and `yp'.
	{
		x = xp;
		y = yp;
	}
	
			
	public abstract void paint (Graphics g);
	// Paint this person in a world
	// with Graphics `g' at its position.

	public void paintIcon(Component c,
                      Graphics g,
                      int x,
                      int y)
	// Paint the icon symbol of this person
	// with graphics `g' on to component `c'
	// at location `x', `y'.
	{
		Graphics2D g2d = (Graphics2D)g;		
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	
		// paint the person as blue filled oval.
		g2d.setColor(color);
		g2d.fillOval(x, y, 12, 12);	
	}
	
	
	public int getIconWidth()
	// get the width of this person's icon symbol width.
	{
		return 12;
	}
	
	
	public int getIconHeight()
	// get the height of this person's icon symbol width.
	{
		return 12;
	}
	
	
	public Color getColor()
	// get the color of this person.
	{
		return color;
	}
	

	public void setColor(Color col)
	// set the color of this person.
	{
		color = col;
	}


	
	public String getName() 
	// get the name of this person.
	{
		return name;
	}
	
	public void setName(String a_name) 
	// set the name of this person to 'a_name'.
	{
		name = a_name;
		buildPaintComponents();
	}


	public PreferenceList getPreferences () 
	// get the preference list of this person.
	{
		return preferences;
	}
	
	public JComponent getNameComponent () 
	// get component to paint the name.
	{
		return name_component;
	}
	
	public JComponent getSymbolComponent () 
	// get component to paint the symbol.
	{
		return symbol_component;
	}
	
	public JComponent getSymbolWithNameComponent () 
	// get component to paint the symbol together with the name.
	{
		return symbol_with_name_component;
	}
	
	public JComponent createSymbolWithNameComponent () 
	// get component to paint the symbol together with the name.
	{		
		// create symbol_with_name_component.
		JLabel symbol_with_name_label = new JLabel(name, (Icon) this, JLabel.CENTER);
		symbol_with_name_label.setVerticalAlignment (JLabel.CENTER);
		symbol_with_name_label.setFont (symbol_with_name_label.getFont().deriveFont(Font.PLAIN));
		symbol_with_name_label.setBackground(Color.black);
		symbol_with_name_label.setForeground(Color.white);
		symbol_with_name_label.setOpaque(true);
		return symbol_with_name_label;			
	}
	
	public boolean occuredAction() 
	// returns true if there was an action in the last step of the person
	// that should stop the world if in stepping mode. 
	{
		return false;
	}
	
	public void extend_preferences (Person p)
	// appends a person to the preferences list.
	{
		preferences.add (p);
	}
	
	public void setTableRow(PersonTableModel table, int row) 
	// Set table and row number where this person is listed.
	{
		ownerTable = table;
		ownerTableRow = row;
	}
	
	public void randomizePreferences ()
	// swaps preferences randomly.
	{
		int count = preferences.size ();
		for (int i = 0; i < count; i++) {
			int r = rnd.nextInt(count - i) + i;
			// swap elements i and r.
			Person p = (Person) preferences.elementAt(r);
			preferences.removeElementAt(r);
			preferences.insertElementAt(p, i);			
		}
	}
	
	public void unrandomizePreferences ()
	// sort preferences (unmix them).
	{
		// create a new preference list.
		int count = preferences.size ();
		PreferenceList new_preferences = new PreferenceList();
		new_preferences.setSize(count);
	
		// put all persons from old preference list
		// to correct position in new preference list.			
		for (int i = 0; i < count; i++) {
			// putElement i to correct position.
			Person p = (Person) preferences.elementAt(i);
			new_preferences.setElementAt(p, p.ownerTableRow);			
		}
		preferences = new_preferences;
	}
	
	public void invertPreferences ()
	// sort preferences (unmix them).
	{
		int count = preferences.size ();
		for (int i = 0; i < count; i++) {
			// move element i to first position.
			Person p = (Person) preferences.elementAt(i);
			preferences.removeElementAt(i);
			preferences.insertElementAt(p, 0);					
		}
	}
	
	
	public abstract boolean isActing();
	// Return true if this person has something to do with the pending action.

	public abstract void step ();
	// let the person do the next step of its movement.
	// (to implement in subclasses)
	
	static boolean is_acting_names_visible;
	// will names of people in action be visualized in world?
	
	static void setNamesVisibleForActingPersons (boolean flag)
	// set if names of people in action will be visible in world.
	{
		is_acting_names_visible = flag;	
	}
	
	static boolean is_married_names_visible;
	// will names of married people be visible in world?
	
	static void setNamesVisibleForMarriedPersons (boolean flag)
	// set if names of married people will be visible in world.
	{
		is_married_names_visible = flag;	
	}

	
			
}
