// ***********************************************************************
// VISUAL MARRIAGE Applet
//
// Visualizes the algorithm for finding a man-optimal stable marriage.
// The simulated algorithm is the one of Gale and Shapley.
//
// by Rolf Bruderer, Oct 2003
// ***********************************************************************

import java.awt.*;
import java.util.*;
import java.util.Date;


public class World 
	// World for visualizing men and women (of type Person)
	// Draws all men and women in a rectangular world.

extends Canvas 

implements Runnable

{


	private Vector persons;
	// All persons in this world.
	
	private Image image_buf;
	// image buffer for double buffering.
	
	private Graphics image_graphics;
	// Graphics for doing double buffering.
	
	private boolean moving;
	// is world in moving state?

	private boolean stepping;
	// is world in stepping state? (waiting after each action?)
	
	private boolean is_alive;
	// is the world alive?
	// which means should the run() method still do something or is the world destroyed?
	
	public World () 
	// initialize world.
	{		
		// initialize canvas.
		setBackground (Color.black);				
		setVisible (true);
		
		// initialize persons (to empty).
		persons = new Vector();
		
		// initialize image_buf.
		image_buf = null;
		image_graphics = null;	
		
		// initialize the world to be alive.
		is_alive = true;
				
	}
	
	public void update (Graphics graph) 
	// update the world image without clearing the background first (avoid flashing).
	{
		paint(graph);
	}
	
	
	public void paint (Graphics graph)
	// Repaint the world and its contents
	// with Graphics `g'.
	{
	
		// Get image for double buffering (if not yet).
		if (image_buf == null) {
			image_buf = createImage (getWidth (), getHeight ());
			image_graphics = image_buf.getGraphics ();
			Graphics2D g2d = (Graphics2D)image_graphics;
			g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		}

		// Empty image with black.
		image_graphics.setColor(Color.black);
		image_graphics.fillRect (0, 0, getWidth(), getHeight());
		
		// Paint persons.
		Person activeWoman = null;
		Enumeration enum_persons = persons.elements();
	    while (enum_persons.hasMoreElements ()) {
			Person p = (Person) (enum_persons.nextElement ());			
			if (p.isActing()) {
				activeWoman = p;
			}
			else {
				p.paint (image_graphics);			
			}
		}
		
		// Paint the acting woman in the end (on top of everything, together with her admirers):
		if (activeWoman!=null) {
			activeWoman.paint (image_graphics);			
		}
		
		// Flush image to screen.
		graph.drawImage (image_buf, 0, 0, getWidth (), getHeight (), this);
	}
	
	public void add (Person a_person) 
	// Add `a_person' to this world.
	{
		persons.add (a_person);	
	}
	
	public void removeAllPersons ()
	// removes all persons from world.
	{
		persons.removeAllElements();
	}
	
	public void run ()
	// Keeps the world running.
	// Always calls update method of each person in the world.
	// Afterwards repaints the world. 	
	// Repeats until it gets somehow interrupted.
	{	

		// set font.
		setFont(getFont().deriveFont(Font.BOLD));
	
		// define the minimum time per step 
		// (so that animation does not become too fast)
		long waittime = 50;		
	
		// run it.
		while (is_alive) {
		
			// get time in milliseconds.
			long time = (new Date()).getTime ();
						
			// make all persons do a step.
			Enumeration enum_persons = persons.elements();
		    	while (enum_persons.hasMoreElements ()) {
				
				// get next person.
				Person p = (Person) (enum_persons.nextElement ());			
				
				// wait on moving state.
				while (!moving) {
					// wait, if world currently not running.
					try { Thread.sleep(waittime); }
					catch (InterruptedException e) {System.out.println ("Error " + e);}
				}
				
				// calculate step.
				p.step ();
				
				// any action occured in last step?
				if (p.occuredAction()) {

					// paint the action.
					repaint ();
					
					// if stepping state then stop after each action,
					if (stepping) {
						stop();
					}						
			
				}
				
			}			
			
			// paint the world.
			repaint ();		
						
			// ensure that it runs not too fast.
			long usedtime = (new Date ()).getTime () - time;
			if (usedtime < waittime) {
				time = waittime - usedtime;
			}
			else {
				time = 1;
			}
			try { Thread.sleep (time); }
			catch (InterruptedException e) {System.out.println ("Error " + e);}
	
			
		}		
	}
	
	public void start() 
	// put world into moving state.
	{	
		moving = true;		
		stepping = false;
	}
	
	public void stop() 
	// put world into stoped state (no moving).
	{	
		moving = false;		
		stepping = false;
	}

	public void step() 
	// put world into stepping state (moving, stop after each step).
	{	
		moving = true;		
		stepping = true;
	}
	
	public void destroy()
	// stops the world and end the running thread.
	{
		is_alive = false;
	}
	
}
